import dtiplayground.dmri.common as common

logger=common.logger
measure_time=common.measure_time
Color=common.Color
get_uuid=common.get_uuid
object_by_id=common.object_by_id
get_timestamp=common.get_timestamp
    